%
%
% Mutual Information (MI) Thresholding
%
% [T1, T2, mi, imT1, imT2, imF, quality, miscore, mii1, mii2] = mutualinfoThreshold(im1, im2, mask, sigma, porder)
%
% Input:
%   im1/im2 : input data pair to be thresholded. Must be the same size
%
%   mask : binary mask the same size as im1/im2. Zero values will be removed from consideration in the data.
%      default = 1
%
%   sigma : smoothing factor for MI surface
%      default = no smoothing
%
%   porder : peak to choose (1=highest peak, 2=second highest, 3=third, etc..  -1=select peak with highest quality)
%      default = 1
%
%
%
% Output:
%
% T1,T2 : threshold for im1/im2
% mi : 256x256 MI surface
% imT1 : im1 thresholded with T1
% imT2 : im2 thresholded with T2
% imF : fusion of imT1 and imT2 using a simple neighbourhood support weighting
% quality : quality score at selected peak
% miscore : MI value at selected peak
% mii1,mii2 : (y,x) position of selected peak on MI surface
%
%
% author: Ciarn  Conaire, oconaire [at] eeng.dcu.ie
%
function [T1, T2, mi, imT1, imT2, imF, quality, miscore, mii1, mii2] = mutualinfoThreshold(im1, im2, mask, sigma, porder)

if ((nargin <= 2) | (prod(size(mask)) < prod(size(im1))))
    % no mask used
    mask=ones(size(im1));
end
if (nargin <= 3)
    sigma = 0.000001; % no smoothing
end
if (nargin <= 4)
    porder = 1; % select top peak
end
if (~isequal(size(im1),size(im2)))
    error('Input data pair are not the same size');
end

min1 = min(makelinear(im1));
min2 = min(makelinear(im2));
max1 = max(makelinear(im1));
max2 = max(makelinear(im2));

im1b = round(255 * (im1-min1) ./ (max1-min1));
im2b = round(255 * (im2-min2) ./ (max2-min2));

[mi,av,bv,cv,dv] = getmutualinfofast(im1b,im2b,mask);

% before filtering, save all the peaks
mi3 = keeplocalmax(mi);
mi3 = mi3 .* (av > ((av+bv).*(av+cv))); % remove results that don't satisfy the quality bounds [0..1]
[mmy,mmx] = find(mi3);
mmz = find(mi3);
allpeaks = [mmx mmy mi3(mmz)];

% smooth mi
gf = GaussianFilter(17,sigma);
mi = filter2(gf,mi) ./ filter2(gf, ones(size(mi)));

% find and sort local maxima greater than the median
tmp = 0*mi+1;
mi2 = keeplocalmax(mi);
mi2 = mi2 .* tmp;
mi2 = mi2 .* (av > ((av+bv).*(av+cv)));

if (porder == -1)
    [h,hc] = hist(makelinear(mi),256);
    miT = hc(rosinthreshold(h));
    mi2 = mi2 .* (mi2 >= miT); % Threshold the MI surface
end

[mmy,mmx] = find(mi2);
mmz = find(mi2);

tmp = [mmx mmy mi2(mmz)];

% If no local maxima are found
if (isempty(tmp))
   % None found that satisfy the quality bounds [0..1]

   % SET: T1, T2, mi, imT1, imT2, imF, quality, miscore, mii1, mii2
   [mii1, mii2] = brightestpoint(mi);
   miscore = max(max(mi));
   quality = -1;
   T1 = (mii1/256) * (max1 - min1) + min1;
   T2 = (mii2/256) * (max2 - min2) + min2;   
   imT1 = (im1 >= T1)+0;
   imT2 = (im2 >= T2)+0;
   imF = (imT1+imT2)/2;
   
   return
end


tmp = -sortrows(-tmp, 3);

% chose which one
if (porder > 0)
    porder = min(porder, size(tmp,1));
    tmp = tmp(porder,1:2);
    mii1 = tmp(2);
    mii2 = tmp(1);
else
    stmp = size(tmp,1);
    plist=tmp;
    %gtmp = rosinthreshold(plist(:,3));
    gtmp = size(plist,1);
    

    if (porder == -1)
        % do Rosin thresholding
        gtmp = max(2,gtmp);
    else
        porder = round(porder);
        plist = plist(1:-porder, :);
    end
    
    bestp = 0; bestq = -1;
    qqlist=[];
    for porder = 1:size(plist,1)
        tmp = plist(porder,1:2);
        mii1 = tmp(2);
        mii2 = tmp(1);
        
        % find nearest peak in original list
        distlist = (allpeaks(:,1)-mii2).^2 + (allpeaks(:,2)-mii1).^2;
        [mmin,mpos] = min(distlist);
        mii1 = allpeaks(mpos,2);
        mii2 = allpeaks(mpos,1);
        
        
        % compute quality
        x=av(mii1,mii2);
        if ((x==0) | (x==1))
           qq=0;
        else
           Lx = (-x*log(x) - (1-x)*log(1-x));
           qq = mi(mii1,mii2) / Lx;
        end
        

        if (qq>bestq)
            bestq=qq;
            bestp = porder;
        end
        qqlist=[qqlist qq];
    end 

    tmp = plist(bestp,1:2);
    mii1 = tmp(2);
    mii2 = tmp(1);    
end


T1 = (mii1/256) * (max1 - min1) + min1;
T2 = (mii2/256) * (max2 - min2) + min2;


imT1 = (im1 >= T1)+0;
imT2 = (im2 >= T2)+0;


K = 5;

filt = ones(3,3);
filt(2,2)=0;
support1 = filter3(filt, imT1);
support1 = (imT1==0).*(support1 <= 8-K) + (imT1==1).*(support1 >= K);
support2 = filter3(filt, imT2);
support2 = (imT2==0).*(support2 <= 8-K) + (imT2==1).*(support2 >= K);
support = (support1 < support2);

support = (support1 == support2);

imF = (imT1~=imT2) .* ((imT1.*support1 + support2.*imT2).*(1-support) + 0.5*support) + imT1.*(imT1==imT2);

miscore = mi(mii1,mii2);


x=av(mii1,mii2);
Lx = (-x*log(x) - (1-x)*log(1-x));
quality = mi(mii1,mii2) / Lx;
        
        
